@@ -4,10 +4,12 @@
 /* Author: Rainer Doemer			first version: 09/24/97 */
 /************************************************************************/
 
-/* last update: 06/15/04 */
+/* last update: 09/26/06 */
 
 /* modification: (most recent first)
  *
+ * 09/26/06 PC  Adjustments for scrc 2.1
+ * 06/03/05 RD	reorganized and renamed global type names
  * 06/15/04 PC  Adjustments for scrc 2.0
  * 08/13/01 RD	switched type of bitvector length to unsigned
  * 05/31/01 RD	eliminated level 2 of SIR API
@@ -98,7 +100,7 @@ float			F_Value;	/* floating constants */
 double			D_Value;
 long double		LD_Value;
 sir_bit			*BIT_Value;	/* bit vector constant */
-string			*CS_Value;	/* character string constant */
+gl_string		*CS_Value;	/* character string constant */
 };
 
 
@@ -131,7 +133,7 @@ SIR_Constant(				/* constructor #4 */
 
 SIR_Constant(				/* constructor #5 */
 	SIR_CONSTTYPE	Type,		/* (the string is consumed!) */
-	string		*Value,
+	gl_string	*Value,
 	unsigned int	Line = 0,
 	sir_fileinfo	*FileInfo = NULL);
 
@@ -168,7 +170,7 @@ static const char *PrettyChar(	/* returns a C-style character */
 	const char	Char);
 
 static void PrettyString(	/* returns a C-style string */
-	string		*Result,
+	gl_string	*Result,
 	const char	*String);
 
 static sir_constant *New(       /* create a new integer constant (#1) */
@@ -176,12 +178,12 @@ static sir_constant *New(       /* create a new integer constant (#1) */
         SIR_CONSTTYPE   CType = SIR_CONST_INT); /* default: integer */
 
 sir_constant *Copy(             /* creates a copy of the constant */
-        BOOL            Strip = TRUE);  /* (returns NULL if SIR_Error) */
+        bool            Strip = true);  /* (returns NULL if SIR_Error) */
 
 void Delete(void);              /* deletes the constant */
 
 const char *Print(		/* return text string of the constants value */
-	BOOL		CppNotation = FALSE);	/* default: SpecC notation */
+	bool		CppNotation = false);	/* default: SpecC notation */
 
 ERROR Compare(                  /* compare this constant with another one (#1)*/
         sir_constant    *Const2,        /* (returns error if not comparable) */
