@@ -4,10 +4,13 @@
 /* Author: Rainer Doemer			first version: 07/02/97 */
 /************************************************************************/
 
-/* last update: 06/15/04 */
+/* last update: 09/26/06 */
 
 /* modifications: (most recent first)
  *
+ * 09/26/06 PC  Adjustments for scrc 2.1
+ * 06/03/05 RD	reorganized and renamed global type names
+ * 12/16/04 RD	added GL_ERROR_FILE_NOT_WRITABLE
  * 06/15/04 PC  Adjustments for scrc 2.0
  * 10/09/02 RD	added 'fsmd', 'rising', 'falling' keywords (SpecC 2.0)
  * 07/11/02 RD	added 'fix', 'buffered', 'signal' keywords (SpecC 2.0)
@@ -89,8 +92,8 @@ static int	SpareMemorySize = 0;
 /*** exported variables *************************************************/
 
 
-ERROR	GL_Error = GL_ERROR_NO_ERROR;	/* last error of this module */
-string	GL_ErrorMessage;
+ERROR		GL_Error = GL_ERROR_NO_ERROR;	/* last error of this module */
+gl_string	GL_ErrorMessage;
 
 int	GL_VerbosityLevel = 0;		/* (published by main module) */
 int	GL_WarningLevel = 0;		/* (published by main module) */
@@ -195,6 +198,10 @@ switch (ErrorNumber)
 			+ GL_SystemErrorMessage */);
      case GL_ERROR_NOT_AT_END_OF_STRING:
 	  return("Not at end of string: excess data found");
+     case GL_ERROR_FILE_NOT_WRITABLE:
+	  return(GL_ErrorMessage.chars()
+		/* "No write permission for file \"%s\""
+			+ GL_SystemErrorMessage */);
     } /* hctiws */
 
 return("Unknown error number (internal)");
@@ -204,7 +211,7 @@ return("Unknown error number (internal)");
 
 const char *GL_SystemErrorMessage(void)
 {
-static string	Message;	/* ('static' ensures valid return value) */
+static gl_string Message;	/* ('static' ensures valid return value) */
 
 if (errno == 0)
    { Message = "";
@@ -767,7 +774,7 @@ const char *GL_GetBaseName(		/* obtains the base name from a path */
 const char	*File;
 char		*Buffer,
 		*BuffPtr;
-static string	BaseName;
+static gl_string BaseName;
 
 File = GL_GetFileFromPath(Path);
 
