@@ -2,6 +2,10 @@
 // bit.h: Bit Vector Library, user level
 // ============================================================================
 // 
+//
+// 09/26/06 PC  Adjustments for scrc 2.1
+// 04/05/06 AG	Added overloaded copy and copy-assignment (for subclassing)
+// 01/12/06 RD	added virtual destructor to template class 'bit'
 // 11/28/01 <gerstl>	Removed constructors from shadow _longlong_
 // 08/10/01 <gerstl>	Introduced len_t, bnd_t; changed shift cnt to unsigned
 // 05/24/01 <gerstl>	Initial version
@@ -44,8 +48,9 @@ public:
   bit()                                                                     { }
   bit(const _bit& c)                                             { ref() = c; }
   bit(const _bitslice& s)                                        { ref() = s; }
-  template<len_t l, bool us> bit(const bit<l,us>& c)             { ref() = c; }
-  template<len_t l, bool us> bit(const _bit_<l,us>& c)           { ref() = c; }
+  bit(const bit& c)                                       { ref() = c; }
+  template<len_t l, bool us> bit(const bit<l,us>& c)      { ref() = c; }
+  template<len_t l, bool us> bit(const _bit_<l,us>& c)    { ref() = c; }
   bit(const char* s)                                             { ref() = s; }
   bit(int i)                                                     { ref() = i; }
   bit(unsigned int i)                                            { ref() = i; }
@@ -53,10 +58,14 @@ public:
   bit(unsigned long i)                                           { ref() = i; }
   bit(LONG_LONG i)                                               { ref() = i; }
   bit(UNSIGNED_LONG_LONG i)                                      { ref() = i; }
+  
+  // Destructor (01/12/06, RD)
+  virtual ~bit()                                                            { }
 
   // Assignments
   bit& operator=(const _bit& o)                    { ref() = o; return put(); }
   bit& operator=(const _bitslice& o)               { ref() = o; return put(); }
+  bit& operator=(const bit& c)              { ref() = c; return put(); }
   template<len_t l, bool us> 
     bit& operator=(const bit<l,us>& c)             { ref() = c; return put(); }
   template<len_t l, bool us> 
@@ -127,24 +136,24 @@ public:
                               
   // Implicit cast to working (rhs) bit vector representation
   operator const _bit() const                { const_cast<bit*>(this)->get();
-                      return _bit(const_cast<_bit::chunk*>(vec), len, usign); }
+                      return _bit(const_cast<_bit::chunk*>(this->vec), len, usign); }
   
   // Create a piece for a bit vector bus
   virtual _bitbus_element* bus_slice(len_t sl, len_t sr) 
-                       { return new _bitbus_element(vec, len, usign, sl, sr); }
+                       { return new _bitbus_element(this->vec, len, usign, sl, sr); }
   _bitbus_element* bus_slice(bnd_t l, bnd_t r, bnd_t sl, bnd_t sr) 
                         { return bus_slice(_BITMAP(l,r,sl), _BITMAP(l,r,sr)); }
 
 protected:
   // Convert to _bit base class
-  _bit ref(void)                              { return _bit(vec, len, usign); }
+  _bit ref(void)                              { return _bit(this->vec, len, usign); }
   const _bit val(void) const                  { return operator const _bit(); }
   
   // Create a slice
   virtual _bitslice slice(_bit::len_t l, _bit::len_t r, bool usgn = usign)
-                             { return _bitslice(vec, len, usign, l, r, usgn); }
+                             { return _bitslice(this->vec, len, usign, l, r, usgn); }
   virtual const _bitslice slice(len_t l, len_t r, bool usgn = usign) const
-   { return _bitslice(const_cast<_bit::chunk*>(vec), len, usign, l, r, usgn); }
+   { return _bitslice(const_cast<_bit::chunk*>(this->vec), len, usign, l, r, usgn); }
   
   // Functions for synchronization to be overloaded by bit busses
   virtual bit& put()                                          { return *this; }
@@ -178,13 +187,13 @@ public:
 protected:  
   // Create a slice
   virtual _bitslice slice(len_t l, len_t r, bool usgn = usign)
-                        { return _bitslice(vec, len, usign, l, r, usgn, lst); }
+                        { return _bitslice(this->vec, len, usign, l, r, usgn, lst); }
   virtual const _bitslice slice(len_t l, len_t r, bool usgn = usign) const
-                            { return _bitslice(const_cast<_bit::chunk*>(vec), 
+                            { return _bitslice(const_cast<_bit::chunk*>(this->vec), 
                     		                len, usign, l, r, usgn, lst); }
   // Sync with original vectors
-  virtual _bitbus& put()                     { lst->put(ref()); return *this; }
-  virtual _bitbus& get()                     { lst->get(ref()); return *this; }
+  virtual _bitbus& put()                     { lst->put(this->ref()); return *this; }
+  virtual _bitbus& get()                     { lst->get(this->ref()); return *this; }
 };
 
 
