@@ -4,10 +4,23 @@
 /* Author: Rainer Doemer			first version: 03/26/01 */
 /************************************************************************/
 
-/* last update: 12/10/01 */
+/* last update: 06/15/04 */
 
 /* modifications: (most recent first)
  *
+ * 06/15/04 PC  Adjustments for scrc 2.0
+ * 03/20/03 RD	call IsNotified() on events instead of direct member access
+ * 01/08/03 RD	added support for self-triggering events (_specc::auto_event)
+ * 12/13/02 RD	added support for asynchronous reset of 'buffered' variables
+ * 12/06/02 RD	applied Notify1List to 'buffered' and 'signal' variables
+ * 12/06/02 RD	distinguish 'signal'/'buffered' variables in separate lists
+ * 11/27/02 RD	added support for signal edges in event lists;
+ *		switched sections 3 and 4 in the scheduler
+ * 10/04/02 RD	changed scheduler to avoid loosing events at exceptions
+ *		(switched the order of notification and exception handling)
+ * 06/26/02 RD	added 'wait' statement with AND semantics (SIM_STATE_WAITAND)
+ * 02/27/02 RD	added support for 'signal' and 'buffered' variables
+ * 01/21/02 RD	added support for non-deterministic (random) scheduler
  * 12/10/01 RD	made return type of ThreadWrapper dependant on thread pkg.
  * 11/21/01 RD	took out default arguments from function definitions
  * 08/01/01 RD	added missing ThreadDelete() call in DeleteChildren()
@@ -69,6 +82,13 @@ void main(void);		/* do nothing */
 static thread_ret ThreadWrapper( /* wrapper around the actual behavior */
 	void		*Arg);
 
+#ifdef SIM_RANDOM_SCHEDULE
+
+static unsigned int Rnd(	/* random number in range [0...Range-1] */
+	unsigned int	Range);
+
+#endif /* SIM_RANDOM_SCHEDULE */
+
 
 /*** internal variables *************************************************/
 
@@ -190,6 +210,7 @@ switch(MyParent->Status)
      case SIM_STATE_NOTIFIED:
      case SIM_STATE_WAITFOR:
      case SIM_STATE_WAIT:
+     case SIM_STATE_WAITAND:
      case SIM_STATE_COMPLETE:
      case SIM_STATE_SUSPENDED:
      case SIM_STATE_SUSPENDED2:
@@ -208,6 +229,23 @@ return(0);	/* dummy result */
 } /* end of ThreadWrapper */
 
 
+#ifdef SIM_RANDOM_SCHEDULE
+
+static unsigned int Rnd(	/* random number in range [0...Range-1] */
+	unsigned int	Range)
+{
+unsigned int	rnd;
+
+rnd = (unsigned int)(drand48() * (double)Range);
+assert(rnd < Range);
+
+return(rnd);
+
+} /* end of Rnd */
+
+#endif /* SIM_RANDOM_SCHEDULE */
+
+
 /*** class implementations **********************************************/
 
 
@@ -359,6 +397,59 @@ Length++;
 } /* end of _specc::queue::Insert */
 
 
+#ifdef SIM_RANDOM_SCHEDULE
+
+void _specc::queue::RandomInsert(	/* insert thread into the queue */
+	thread		*Thread)	/* at random position           */
+{
+thread		*Prev;
+unsigned int	RndPos;
+
+assert(Thread != NULL);		/* must exist */
+assert(Thread->Next == NULL);	/* must not be in any queue */
+assert(Thread->Prev == NULL);
+assert(Thread->Queue == NULL);
+
+if (First)
+   { assert(Last != NULL);
+     RndPos = Rnd(Length+1);
+     if (RndPos > 0)
+	{ Prev = First;		/* search predecessor */
+	  while(0 <-- RndPos)
+	     { Prev = Prev->Next;
+	      } /* elihw */
+	  assert(Prev != NULL);
+	  Thread->Next = Prev->Next;	/* insert here */
+	  Thread->Prev = Prev;
+	  if (Prev->Next)
+	     { Prev->Next->Prev = Thread;
+	      } /* fi */
+	  else
+	     { assert(Last == Prev);
+	       Last = Thread;
+	      } /* esle */
+	  Prev->Next = Thread;
+	 } /* fi */
+     else	/* insert at the very front (position 0) */
+	{ Thread->Next = First;
+	  First->Prev = Thread;
+	  First = Thread;
+	 } /* esle */
+    } /* fi */
+else	/* insert as the only element, no choice of position */
+   { assert(Last == NULL);
+     First = Thread;
+     Last = Thread;
+    } /* esle */
+
+Thread->Queue = this;
+Length++;
+
+} /* end of _specc::queue::RandomInsert */
+
+#endif /* SIM_RANDOM_SCHEDULE */
+
+
 void _specc::queue::Remove(	/* remove the thread from this queue */
 	thread	*Thread)
 {
@@ -481,6 +572,9 @@ delete Exceptions;
 void _specc::thread::Start(	/* initialize thread usage */
 	void)
 {
+#ifdef SIM_RANDOM_SCHEDULE
+unsigned short	seed16v[3] = SIM_RANDOM_SEED;
+#endif /* SIM_RANDOM_SCHEDULE */
 
 #ifdef SIM_THREAD_DEBUG
 puts("SIM_THREAD_DEBUG: Initializing simulation time...");
@@ -498,6 +592,13 @@ assert(SuspendQueue.IsEmpty());
 assert(NotifiedEvents == NULL);	/* event lists must be empty */
 assert(Notify1List == NULL);
 
+#ifdef SIM_RANDOM_SCHEDULE
+#ifdef SIM_THREAD_DEBUG
+puts("SIM_THREAD_DEBUG: Seeding random number generator...");
+#endif /* SIM_THREAD_DEBUG */
+seed48(seed16v);
+#endif /* SIM_RANDOM_SCHEDULE */
+
 #ifdef SIM_THREAD_DEBUG
 puts("SIM_THREAD_DEBUG: Initializing thread package...");
 #endif /* SIM_THREAD_DEBUG */
@@ -527,10 +628,7 @@ assert(NotifiedQueue.IsEmpty());
 assert(TryQueue.IsEmpty());
 assert(SuspendQueue.IsEmpty());
 
-if (NotifiedEvents)		/* clean up notification lists */
-   { NotifiedEvents->Reset();
-     NotifiedEvents = NULL;
-    } /* fi */
+_specc::event::UnNotifyAll();	/* clean up notification lists */
 delete Notify1List;
 Notify1List = NULL;
 
@@ -817,24 +915,43 @@ thread		*NextThread,
 		*Succ,
 		*SelectedThread,
 		*NextThreadPred;
-event_ptr	*EventPtr;
+event		*SelectedEvent;
+auto_event	*AutoEvent,
+		*AutoEventSucc;
+event_ptr	*EventPtr,
+		**EventPtrPtr;
 event_list	*Notify1,
 		**Notify1Ptr;
 exception	*Exception;
-sim_time	NextTime;
+sim_time	NextTime,
+		AutoEventTimeOut;
+buffered_base	*BuffVar;
+bool		UpdateBuffVar;
+#ifdef SIM_NOTIFYONE_SELECT_RANDOM_WAITING_THREAD
+unsigned int	Candidates,
+		Candidate,
+		SelectedCandidate;
+#endif /* SIM_NOTIFYONE_SELECT_RANDOM_WAITING_THREAD */
 
 assert(CurrentThread == this);		/* this is the current thread */
 assert(! IsInQueue(&ReadyQueue));	/* it is not in the ready queue */
 assert(  (Status == SIM_STATE_RUNNING)	/* it is in state RUNNING or */
        ||(Status == SIM_STATE_WAITFOR)	/* was RUNNING just before   */
        ||(Status == SIM_STATE_WAIT)
+       ||(Status == SIM_STATE_WAITAND)
        ||(Status == SIM_STATE_JOIN)
        ||(Status == SIM_STATE_COMPLETE)
        ||(Status == SIM_STATE_TRYING));
 
+AutoEventTimeOut = 0;		/* not in auto-event loop yet */
+
+#ifdef GL_DELTA_SUPPORT
+ScheduleTime = getRealTime();	/* create real time stamp for 'delta' */
+#endif /* GL_DELTA_SUPPORT */
 
 	/*** scheduler part 1: handle the ready queue ***/
 
+	/*** scheduler section 1: handle the ready queue ***/
 
 if (Status == SIM_STATE_RUNNING)	/* possible context switch: */
    {
@@ -854,11 +971,167 @@ if ((NextThread = ReadyQueue.First))	/* any thread ready to run? */
     } /* fi */
 
 
-	/*** scheduler part 2: handle any exceptions ***/
-
+DeliverNotifiedEvents:		//// BEGIN AUTO-EVENT LOOP ////
 
 assert(ReadyQueue.IsEmpty());	/* ready queue is empty here */
 
+
+	/*** scheduler section 2: handle any notifications of 'wait' ***/
+
+
+if (Notify1List)	/* perform notification handling for 'notifyone' */
+   { Notify1 = Notify1List;
+     while(Notify1)
+	{ SelectedThread = NULL;
+	  SelectedEvent = NULL;
+#ifdef SIM_NOTIFYONE_SELECT_RANDOM_WAITING_THREAD
+	  Candidates = 0;
+	  Thread = WaitQueue.First;
+	  while(Thread)
+	     { if (Notify1->Triggers(Thread->EventList))
+		  { Candidates++;
+		   } /* fi */
+	       Thread = Thread->Next;
+	      } /* elihw */
+	  Thread = NotifiedQueue.First;
+	  while(Thread)
+	     { if (Notify1->Triggers(Thread->EventList))
+		  { Candidates++;
+		   } /* fi */
+	       Thread = Thread->Next;
+	      } /* elihw */
+	  if (Candidates > 0)
+	     { Thread = WaitQueue.First;
+	       Candidate = 0;
+	       SelectedCandidate = Rnd(Candidates);
+	       while(Thread)
+		  { if ((SelectedEvent = Notify1->Triggers(Thread->EventList)))
+		       { if (Candidate++ == SelectedCandidate)
+			    { SelectedThread = Thread;
+			      break;
+			     } /* fi */
+			} /* fi */
+		    Thread = Thread->Next;
+		   } /* elihw */
+	       /* no need to search NotifiedQueue since nothing will change */
+	      } /* fi */
+#endif /* SIM_NOTIFYONE_SELECT_RANDOM_WAITING_THREAD */
+#ifdef SIM_NOTIFYONE_SELECT_FIRST_WAITING_THREAD
+	  Thread = WaitQueue.First;
+	  while(Thread)
+	     { if ((SelectedEvent = Notify1->Triggers(Thread->EventList)))
+		  { SelectedThread = Thread;
+		    break;
+		   } /* fi */
+	       Thread = Thread->Next;
+	      } /* elihw */
+#endif /* SIM_NOTIFYONE_SELECT_FIRST_WAITING_THREAD */
+#ifdef SIM_NOTIFYONE_SELECT_LAST_WAITING_THREAD
+	  Thread = WaitQueue.Last;
+	  while(Thread)
+	     { if ((SelectedEvent = Notify1->Triggers(Thread->EventList)))
+		  { SelectedThread = Thread;
+		    break;
+		   } /* fi */
+	       Thread = Thread->Prev;
+	      } /* elihw */
+#endif /* SIM_NOTIFYONE_SELECT_LAST_WAITING_THREAD */
+	  if (SelectedThread)
+	     { assert(SelectedEvent != NULL);
+	       assert(  (SelectedThread->Status == SIM_STATE_WAIT)
+		      ||(SelectedThread->Status == SIM_STATE_WAITAND));
+	       if (SelectedThread->Status == SIM_STATE_WAIT)
+		  { SelectedThread->Status = SIM_STATE_NOTIFIED;
+		    WaitQueue.Remove(SelectedThread);
+		    NotifiedQueue.Append(SelectedThread);
+		   } /* fi */
+	       else
+		  { assert(SelectedThread->Status == SIM_STATE_WAITAND);
+		    EventPtrPtr = &SelectedThread->EventList;
+		    while((EventPtr = *EventPtrPtr))
+		       { if (  (EventPtr->Event == SelectedEvent)
+			     &&(EventPtr->Event->IsValidEdge(EventPtr->Edge)))
+			    { *EventPtrPtr = EventPtr->Next;
+			      EventPtr->Next = NULL; /* take notified events */
+			      delete EventPtr;	/* out from the waiting list */
+			     } /* fi */
+			 else
+			    { EventPtrPtr = &EventPtr->Next;
+			     } /* esle */
+			} /* elihw */
+		    if (! SelectedThread->EventList)	/* list empty? */
+		       { SelectedThread->Status = SIM_STATE_NOTIFIED;
+			 WaitQueue.Remove(SelectedThread);
+			 NotifiedQueue.Append(SelectedThread);
+			} /* fi */
+		   } /* esle */
+	      } /* fi */
+	  else
+	     { /* we could search further in the NotifiedQueue, */
+	       /* but since those threads are already notified, */
+	       /* nothing would change; so, we stop here;	*/
+	      } /* esle */
+	  Notify1 = Notify1->Next;
+	 } /* elihw */
+    } /* fi */
+
+if (NotifiedEvents)	/* perform notification handling for 'notify' */
+   { Thread = WaitQueue.First;
+     while(Thread)
+	{ Succ = Thread->Next;
+	  assert(  (Thread->Status == SIM_STATE_WAIT)
+		 ||(Thread->Status == SIM_STATE_WAITAND));
+	  if (Thread->Status == SIM_STATE_WAIT)
+	     { EventPtr = Thread->EventList;
+	       while(EventPtr)
+		  { if (  (EventPtr->Event->IsNotified())
+			&&(EventPtr->Event->IsValidEdge(EventPtr->Edge)))
+		       { Thread->Status = SIM_STATE_NOTIFIED;
+			 WaitQueue.Remove(Thread);
+			 NotifiedQueue.Append(Thread);
+			 break;
+			} /* fi */
+		    EventPtr = EventPtr->Next;
+		   } /* elihw */
+	      } /* fi */
+	  else
+	     { assert(Thread->Status == SIM_STATE_WAITAND);
+	       EventPtrPtr = &Thread->EventList;
+	       while((EventPtr = *EventPtrPtr))
+		  { if (  (EventPtr->Event->IsNotified())
+			&&(EventPtr->Event->IsValidEdge(EventPtr->Edge)))
+		       { *EventPtrPtr = EventPtr->Next;
+			 EventPtr->Next = NULL;	/* take notified events out  */
+			 delete EventPtr;	/* from the AND-waiting list */
+			} /* fi */
+		    else
+		       { EventPtrPtr = &EventPtr->Next;
+			} /* esle */
+		   } /* elihw */
+	       if (! Thread->EventList)	/* list empty? all events notified? */
+		  { Thread->Status = SIM_STATE_NOTIFIED;
+		    WaitQueue.Remove(Thread);
+		    NotifiedQueue.Append(Thread);
+		   } /* fi */
+	      } /* esle */
+	  Thread = Succ;
+	 } /* elihw */
+    } /* fi */
+
+Thread = NotifiedQueue.First;	/* set all notified threads ready to run */
+while(Thread)
+   { Succ = Thread->Next;
+     assert(Thread->Status == SIM_STATE_NOTIFIED);
+     Thread->Status = SIM_STATE_READY;
+     NotifiedQueue.Remove(Thread);
+     ReadyQueue.ADD(Thread);
+     Thread = Succ;
+    } /* elihw */
+
+
+	/*** scheduler section 3: handle any exceptions ***/
+
+
 Thread = TryQueue.First;/* (note: the order of the try queue obeys the */
 while(Thread)		/*        hierarchical priority of exceptions) */
    { assert(Thread->Status == SIM_STATE_TRYING);
@@ -899,84 +1172,92 @@ while(Thread)		/*        hierarchical priority of exceptions) */
     } /* elihw */
 
 
-	/*** scheduler part 3: handle any notifications ***/
+	/*** scheduler section 4: update 'buffered'/'signal' variables ***/
 
 
-if (Notify1List)	/* perform notification handling for 'notifyone' */
-   { Notify1 = Notify1List;
-     while(Notify1)
-	{ SelectedThread = NULL;
-#ifdef SIM_NOTIFYONE_SELECT_FIRST_WAITING_THREAD
-	  Thread = WaitQueue.First;
-	  while(Thread)
-	     { if (Notify1->Triggers(Thread->EventList))
-		  { SelectedThread = Thread;
-		    break;
+// note that the order "buffered variables first, signals last"
+// is important here, because buffered variables can depend on
+// 'rising' or 'falling' signals (but not vice versa)
+
+if (  (NotifiedEvents)
+    ||(Notify1List))
+   { BuffVar = BufferedVars;	/* perform update for 'buffered' variables */
+     while(BuffVar)
+	{ if (  (BuffVar->ResetSignal)	/* handle asynchronous reset */
+	      &&(BuffVar->ResetSignal->ResetIsActive(BuffVar->ResetActiveHi)))
+	     { /* note: we don't need to check the Notify1List here, */
+	       /* because 'notifyone' on a 'signal' has no effect on */
+	       /* whether the signal value will change or not!       */
+	       if (BuffVar->ResetSignal->IsNotified())
+		  { BuffVar->reset();
 		   } /* fi */
-	       Thread = Thread->Next;
-	      } /* elihw */
-#endif /* SIM_NOTIFYONE_SELECT_FIRST_WAITING_THREAD */
-#ifdef SIM_NOTIFYONE_SELECT_LAST_WAITING_THREAD
-	  Thread = WaitQueue.Last;
-	  while(Thread)
-	     { if (Notify1->Triggers(Thread->EventList))
-		  { SelectedThread = Thread;
-		    break;
-		   } /* fi */
-	       Thread = Thread->Prev;
-	      } /* elihw */
-#endif /* SIM_NOTIFYONE_SELECT_LAST_WAITING_THREAD */
-	  if (SelectedThread)
-	     { assert(SelectedThread->Status == SIM_STATE_WAIT);
-	       SelectedThread->Status = SIM_STATE_NOTIFIED;
-	       WaitQueue.Remove(SelectedThread);
-	       NotifiedQueue.Append(SelectedThread);
+	       BuffVar = BuffVar->Next;
+	       continue;	/* skip the regular update part! */
 	      } /* fi */
-	  else
-	     { /* we could search further in the NotifiedQueue, */
-	       /* but since those threads are already notified, */
-	       /* nothing would change; so, we stop here;	*/
-	      } /* esle */
-	  Notify1 = Notify1->Next;
+	  UpdateBuffVar = false;
+	  if (NotifiedEvents)
+	     { EventPtr = BuffVar->UpdateEvents;
+	       while(EventPtr)
+		  { if (  (EventPtr->Event->IsNotified())
+			&&(EventPtr->Event->IsValidEdge(EventPtr->Edge)))
+		       { UpdateBuffVar = true;
+			 break;
+			} /* fi */
+		    EventPtr = EventPtr->Next;
+		   } /* elihw */
+	      } /* fi */
+	  if (  (! UpdateBuffVar)
+	      &&(Notify1List))
+	     { Notify1Ptr = &Notify1List;
+	       Notify1 = Notify1List;
+	       while(Notify1)
+		  { if (Notify1->Triggers(BuffVar->UpdateEvents))
+		       { UpdateBuffVar = true;
+			 *Notify1Ptr = Notify1->Next; /* consume 'notifyone' */
+			 Notify1->Next = NULL;
+			 delete Notify1;
+			 break;
+			} /* fi */
+		    Notify1Ptr = &Notify1->Next;
+		    Notify1 = Notify1->Next;
+		   } /* elihw */
+	      } /* fi */
+	  if (UpdateBuffVar)
+	     { BuffVar->update();	/* update the buffer data */
+	      } /* fi */
+	  BuffVar = BuffVar->Next;
+	 } /* elihw */
+
+     BuffVar = SignalVars;	/* perform update for 'signal' variables */
+     while(BuffVar)
+	{ /* note: no asynchronous reset handling needed for signals! */
+	  assert(BuffVar->ResetSignal == NULL);
+	  /* note: also, no checking of the Notify1List for signals,  */
+	  /*       because 'notifyone' on a 'signal' has no effect on */
+	  /*       whether the signal value will change or not!       */
+	  if (NotifiedEvents)
+	     { EventPtr = BuffVar->UpdateEvents;
+	       while(EventPtr)
+		  { if (EventPtr->Event->IsNotified())
+		       { assert(EventPtr->Event->IsValidEdge( // is a signal!
+							EventPtr->Edge));
+			 BuffVar->update();	/* update the signal data */
+			 break;
+			} /* fi */
+		    EventPtr = EventPtr->Next;
+		   } /* elihw */
+	      } /* fi */
+	  BuffVar = BuffVar->Next;
 	 } /* elihw */
-     delete Notify1List;	/* reset all 'notifyone' events */
-     Notify1List = NULL;
     } /* fi */
 
-if (NotifiedEvents)	/* perform notification handling for 'notify' */
-   { Thread = WaitQueue.First;
-     while(Thread)
-	{ Succ = Thread->Next;
-	  EventPtr = Thread->EventList;
-	  while(EventPtr)
-	     { if (EventPtr->Event->Notified)
-		  { assert(Thread->Status == SIM_STATE_WAIT);
-		    Thread->Status = SIM_STATE_NOTIFIED;
-		    WaitQueue.Remove(Thread);
-		    NotifiedQueue.Append(Thread);
-		    break;
-		   } /* fi */
-	       EventPtr = EventPtr->Next;
-	      } /* elihw */
-	  Thread = Succ;
-	 } /* elihw */
-     NotifiedEvents->Reset();	/* reset all 'notify' events */
-     NotifiedEvents = NULL;
-    } /* fi */
 
-Thread = NotifiedQueue.First;	/* set all notified threads ready to run */
-while(Thread)
-   { Succ = Thread->Next;
-     assert(Thread->Status == SIM_STATE_NOTIFIED);
-     Thread->Status = SIM_STATE_READY;
-     NotifiedQueue.Remove(Thread);
-     ReadyQueue.ADD(Thread);
-     Thread = Succ;
-    } /* elihw */
+	/*** scheduler section 5: handle the synchronization cycle ***/
 
 
-	/*** scheduler part 4: handle synchronization cycle ***/
-
+_specc::event::UnNotifyAll();	/* reset all 'notify' events */
+delete Notify1List;		/* reset all 'notifyone' events */
+Notify1List = NULL;
 
 if ((NextThread = ReadyQueue.First))	/* any thread ready to run? */
    { ReadyQueue.Remove(NextThread);
@@ -985,18 +1266,28 @@ if ((NextThread = ReadyQueue.First))	/* any thread ready to run? */
      return;
     } /* fi */
 
-
-	/*** scheduler part 5: advance simulation time ***/
-
-
 assert(ReadyQueue.IsEmpty());	/* ready queue is empty here */
 
-if (! WaitforQueue.IsEmpty())	/* anybody waiting for time advance? */
+
+	/*** scheduler section 6: advance the simulation time ***/
+
+
+if (  (!WaitforQueue.IsEmpty())	/* anybody waiting for time advance? */
+    &&(  (! FirstAutoEvent)	/* and no auto-event before that?    */
+       ||(FirstAutoEvent->TriggerTime >=
+		WaitforQueue.First->TimeStamp)))
    { Thread = WaitforQueue.First;
      assert(Thread != NULL);
      NextTime = Thread->TimeStamp;
      assert(NextTime >= CurrentTime);	/* never go backwards in time! */
      CurrentTime = NextTime;		/* advance the simulation time */
+     AutoEvent = FirstAutoEvent;
+     while(  (AutoEvent)
+	   &&(AutoEvent->TriggerTime == CurrentTime))
+	  { AutoEventSucc = AutoEvent->Succ;
+	    AutoEvent->TriggerAndUpdate();
+	    AutoEvent = AutoEventSucc;
+	   } /* elihw */
      do { Succ = Thread->Next;
 	  WaitforQueue.Remove(Thread);	/* enable threads that wake up now */
 	  Thread->TimeStamp = 0;
@@ -1014,8 +1305,66 @@ if (! WaitforQueue.IsEmpty())	/* anybody waiting for time advance? */
      return;
     } /* fi */
 
+if ((AutoEvent = FirstAutoEvent))	/* handle the next auto-event */
+   { if (AutoEventTimeOut == 0)
+	{ assert(LastAutoEvent != NULL);
+	  AutoEventTimeOut = LastAutoEvent->TriggerTime;
+	  assert(AutoEventTimeOut >= CurrentTime);
+	 } /* fi */
 
-	/*** scheduler part 6: handle deadlocks ***/
+     NextTime = AutoEvent->TriggerTime;
+     assert(NextTime >= CurrentTime);	/* never go backwards in time! */
+     CurrentTime = NextTime;		/* advance the simulation time */
+     while(  (AutoEvent)
+	   &&(AutoEvent->TriggerTime == CurrentTime))
+	  { AutoEventSucc = AutoEvent->Succ;
+	    AutoEvent->TriggerAndUpdate();
+	    AutoEvent = AutoEventSucc;
+	   } /* elihw */
+
+     // note:
+     // since self-triggering events are handled completely by the simulation
+     // engine (no user code is executed) and are typically running forever,
+     // there is a good chance that at the "end" of a design simulation all
+     // user code is completed but some auto-events are still active;
+     // in this case, the simulation engine would loop forever in the
+     // scheduler, doing nothing for the user;
+     // so, we'll handle this case as a special deadlock situation:
+
+     if (  (WaitforQueue.IsEmpty())
+	 &&(CurrentTime > AutoEventTimeOut))
+	{ /* deadlock situation! */
+	  assert(ReadyQueue.IsEmpty());
+	  assert(WaitforQueue.IsEmpty());
+	  assert(NotifiedQueue.IsEmpty());
+	  // TryQueue may or may not be empty
+	  // WaitQueue may or may not be empty
+	  // SuspendQueue may or may not be empty
+	  assert(FirstAutoEvent != NULL);	// not empty, but ineffective!
+
+	  fprintf(stderr,
+		"\n"
+		"libsim: Deadlock detected!\n"
+		"libsim: Time %s, 0 threads ready, %u waiting, %u suspended.\n"
+		"libsim: Exiting.\n",
+		time2str(CurrentTime), WaitQueue.Length, SuspendQueue.Length);
+
+	  exit(RESULT_DEADLOCK);
+	 } /* fi */
+     /* else: continue running! */
+
+     // note:
+     // normally, we would now call this scheduler again and then return
+     // (i.e. use "Schedule(); return;");
+     // however, that would be an unnecessary recursion of potentially
+     // high depth; in order to avoid this (and keep the stack usage low),
+     // we use plain iteration instead, by use of a 'goto' statement
+
+     goto DeliverNotifiedEvents;	//// END AUTO-EVENT LOOP ////
+    } /* fi */
+
+
+	/*** scheduler section 7: handle deadlocks ***/
 
 
 assert(ReadyQueue.IsEmpty());		/* ready queue is empty here */
@@ -1024,6 +1373,7 @@ assert(NotifiedQueue.IsEmpty());	/* notify queue is empty anyways */
 // TryQueue may or may not be empty
 // WaitQueue may or may not be empty
 // SuspendQueue may or may not be empty
+assert(FirstAutoEvent == NULL);		/* empty as well */
 
 /* note: typically, there is at least one thread stuck	*/
 /*       in the WaitQueue or in the SuspendQueue;	*/
